@@ -1,19 +1,18 @@
 #!/usr/bin/python
 """ print and edit ovmf varstore files """
-import os
 import sys
 import struct
 import pprint
-import hashlib
 import optparse
 import datetime
 import pkg_resources
 
 from cryptography import x509
-from cryptography.hazmat.primitives import serialization
 
 from ovmfctl.efi import guids
 from ovmfctl.efi import ucs16
+from ovmfctl.efi import devpath
+from ovmfctl.efi import siglist
 
 ##################################################################################################
 # constants
@@ -69,25 +68,13 @@ vars_settings = {
 }
 
 var_template = {
-    'ascii_guid' : 'FIXME',
-    'ascii_name' : 'FIXME',
     'guid'       : b'FIXME',
     'name'       : b'FIXME',
     'attr'       : 0,
     'count'      : 0,
     'pkidx'      : 0,
     'data'       : b'',
-    'time'       : {
-        'year'   : 0,
-        'month'  : 0,
-        'day'    : 0,
-        'hour'   : 0,
-        'min'    : 0,
-        'ns'     : 0,
-        'sec'    : 0,
-        'tz'     : 0,
-        'dl'     : 0,
-    }
+    'time'       : None,
 }
 
 
@@ -97,66 +84,12 @@ var_template = {
 def parse_time(data, offset):
     (year, month, day, hour, minute, second, ns, tz, dl) = \
         struct.unpack_from("=HBBBBBxLhBx", data, offset)
-    time = {
-        'year'  : year,
-        'month' : month,
-        'day'   : day,
-        'hour'  : hour,
-        'min'   : minute,
-        'sec'   : second,
-        'ns'    : ns,
-        'tz'    : tz,
-        'dl'    : dl,
-    }
-    return time
+    if year == 0:
+        return None
+    return datetime.datetime(year, month, day,
+                             hour, minute, second, int(ns / 1000))
 
-def extract_cert(var, owner, cert):
-    cn = cert.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0]
-    filename = var['ascii_name']
-    filename += '-' + owner
-    filename += '-' + "".join(x for x in cn.value if x.isalnum())
-    filename += ".pem"
-    if os.path.exists(filename):
-        print(f"# WARNNG: exists: {filename}")
-        return
-    print(f"# writing: {filename}")
-    with open(filename, "wb") as f:
-        f.write(cert.public_bytes(serialization.Encoding.PEM))
-
-def parse_sigs(var, extract):
-    data = var['data']
-    pos = 0
-    var['siglists'] = []
-    while pos < len(data):
-        guid = guids.parse(data, pos)
-        (lsize, hsize, ssize) = struct.unpack_from("=LLL", data, pos + 16)
-        siglist = data[ pos + 16 + 12 + hsize : pos+lsize ]
-        sigs = []
-        spos = 0
-        while spos < len(siglist):
-            owner = siglist[ spos : spos + 16 ]
-            sdata = siglist[ spos + 16 : spos + ssize ]
-            sig = {
-                'ascii_guid' : guids.parse(owner, 0),
-                'guid'       : owner,
-                'data'       : sdata,
-            }
-            if guid == guids.EfiCertX509:
-                sig['x509'] = x509.load_der_x509_certificate(sdata)
-                if extract:
-                    extract_cert(var, sig['ascii_guid'], sig['x509'])
-            sigs.append(sig)
-            spos += ssize
-
-        var['siglists'].append({
-            'ascii_guid' : guid,
-            'guid'       : data[ pos : pos + 16 ],
-            'header'     : data[ pos + 16 + 12 : pos + 16 + 12 + hsize ],
-            'sigs'       : sigs,
-        })
-        pos += lsize
-
-def parse_vars(data, start, end, extract):
+def parse_vars(data, start, end):
     pos = start
     varlist = {}
     while pos < end:
@@ -171,35 +104,29 @@ def parse_vars(data, start, end, extract):
                 'count' : count,
                 'pkidx' : pk,
             }
-            var['guid'] = data[pos + 44 :
-                               pos + 44 + 16]
-            var['name'] = data[pos + 44 + 16 :
-                               pos + 44 + 16 + nsize]
+            var['time'] = parse_time(data, pos + 16)
+
+            var['guid'] = guids.parse_bin(data, pos + 44)
+            var['name'] = ucs16.from_ucs16(data, pos + 44 + 16)
             var['data'] = data[pos + 44 + 16 + nsize :
                                pos + 44 + 16 + nsize + dsize]
 
-            var['time'] = parse_time(data, pos + 16)
-            var['ascii_guid'] = guids.parse(var['guid'], 0)
-            var['ascii_name'] = ucs16.from_ucs16(var['name'], 0)
-            varlist[var['ascii_name']] = var
+            name = str(var['name'])
+            varlist[name] = var
 
-            if (var['ascii_name'] == "PK"  or
-                var['ascii_name'] == "KEK" or
-                var['ascii_name'] == "db"  or
-                var['ascii_name'] == "dbx" or
-                var['ascii_name'] == "MokList"):
-                parse_sigs(var, extract)
+            if name in ("PK", "KEK", "db", "dbx", "MokList"):
+                var['sigdb'] = siglist.EfiSigDB(var['data'])
 
         pos = pos + 44 + 16 + nsize + dsize
         pos = (pos + 3) & ~3 # align
     return varlist
 
 def parse_varstore(file, data, start):
-    guid = guids.parse(data, start)
+    guid = guids.parse_bin(data, start)
     (size, storefmt, state) = struct.unpack_from("=LBB", data, start + 16)
     print(f'varstore={guids.name(guid)} size=0x{size:x} '
           f'format=0x{storefmt:x} state=0x{state:x}')
-    if guid != guids.AuthVars:
+    if str(guid) != guids.AuthVars:
         print(f"ERROR: {file}: unknown varstore guid")
         sys.exit(1)
     if storefmt != 0x5a:
@@ -211,7 +138,7 @@ def parse_varstore(file, data, start):
     return (start + 16 + 12, start + size)
 
 def parse_volume(file, data):
-    guid = guids.parse(data, 16)
+    guid = guids.parse_bin(data, 16)
     (vlen, sig, attr, hlen, csum, xoff, rev, blocks, blksize) = \
         struct.unpack_from("=QLLHHHxBLL", data, 32)
     print(f'vol={guids.name(guid)} vlen=0x{vlen:x} rev={rev} '
@@ -219,7 +146,7 @@ def parse_volume(file, data):
     if sig != 0x4856465f:
         print(f"ERROR: {file}: not a firmware volume")
         sys.exit(1)
-    if guid != guids.NvData:
+    if str(guid) != guids.NvData:
         print(f"ERROR: {file}: not a variable store")
         sys.exit(1)
     return parse_varstore(file, data, hlen)
@@ -263,8 +190,9 @@ def print_bool(var):
 def print_boot_entry(var):
     (attr, pathsize) = struct.unpack_from('=LH', var['data'])
     name = ucs16.from_ucs16(var['data'], 6)
-    pathoffset = ucs16.get_size(var['data'], 6) + 6
-    print(f'    boot entry: name={name} pathoffset=0x{pathoffset:x}')
+    pathoffset = name.size() + 6
+    devicepath = devpath.DevicePath(var['data'][pathoffset: pathoffset + pathsize])
+    print(f'    boot entry: name={name} devicepath={devicepath}')
 
 def print_boot_list(var):
     bootlist = []
@@ -277,14 +205,13 @@ def print_boot_list(var):
 def print_ascii(var):
     print(f"    string: {var['data'].decode()}")
 
-def print_siglists(var):
-    for item in var['siglists']:
-        name = guids.name(item['ascii_guid'])
-        count = len(item['sigs'])
+def print_sigdb(var):
+    for item in var['sigdb']:
+        name = guids.name(item.guid)
+        count = len(item)
         print(f'    list type={name} count={count}')
-        cert = item['sigs'][0].get('x509')
-        if cert:
-            cn = cert.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0]
+        if item.x509:
+            cn = item.x509.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0]
             print(f'      x509 CN={cn.value}')
 
 print_funcs = {
@@ -309,15 +236,15 @@ print_funcs = {
 }
 
 def print_var(var, verbose, hexdump):
-    name = var['ascii_name']
-    gname = guids.name(var['ascii_guid'])
+    name = str(var['name'])
+    gname = guids.name(var['guid'])
     size = len(var['data'])
     print(f'  - name={name} guid={gname} size={size}')
-    pfunc = print_funcs.get(var['ascii_name'])
+    pfunc = print_funcs.get(name)
     if pfunc:
         pfunc(var)
-    if var.get('siglists'):
-        print_siglists(var)
+    if var.get('sigdb'):
+        print_sigdb(var)
     if verbose:
         print("----- raw -----")
         pprint.pprint(var)
@@ -335,33 +262,20 @@ def print_vars(varlist, verbose, hexdump):
 ##################################################################################################
 # write vars
 
-def update_data_from_siglists(var):
-    siglists = var.get('siglists')
-    if not siglists:
+def update_data_from_sigdb(var):
+    sigdb = var.get('sigdb')
+    if not sigdb:
         return
-    blob = b''
-    for siglist in siglists:
-        sigs = b''
-        count = 0
-        for sig in siglist['sigs']:
-            sigs += sig['guid']
-            sigs += sig['data']
-            count += 1
-        blob += siglist['guid']
-        blob += struct.pack("=LLL",
-                            16 + 12 + len(siglist['header']) + len(sigs),
-                            len(siglist['header']),
-                            int(len(sigs) / count))
-        blob += siglist['header']
-        blob += sigs
-    var['data'] = blob
+    var['data'] = bytes(sigdb)
 
 def write_time(time):
-    blob = struct.pack("=HBBBBBxLhBx",
-                       time['year'], time['month'], time['day'],
-                       time['hour'], time['min'], time['sec'],
-                       time['ns'], time['tz'], time['dl'])
-    return blob
+    if time is None:
+        return b'\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0'
+    return struct.pack("=HBBBBBxLhBx",
+                       time.year, time.month, time.day,
+                       time.hour, time.minute, time.second,
+                       time.microsecond * 1000,
+                       0, 0)
 
 def write_var(var):
     blob = struct.pack("=HBxLQ",
@@ -371,10 +285,10 @@ def write_var(var):
     blob += write_time(var['time'])
     blob += struct.pack("=LLL",
                         var['pkidx'],
-                        len(var['name']),
+                        var['name'].size(),
                         len(var['data']))
-    blob += var['guid']
-    blob += var['name']
+    blob += var['guid'].bytes_le
+    blob += bytes(var['name'])
     blob += var['data']
     while len(blob) & 3:
         blob += b'\0'
@@ -391,13 +305,7 @@ def vars_delete(varlist, delete):
 def var_update_time(var):
     if not var['attr'] & EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS:
         return
-    now = datetime.datetime.now(datetime.timezone.utc)
-    var['time']['year']  = now.year
-    var['time']['month'] = now.month
-    var['time']['day']   = now.day
-    var['time']['hour']  = now.hour
-    var['time']['min']   = now.minute
-    var['time']['sec']   = now.second
+    var['time'] = datetime.datetime.now(datetime.timezone.utc)
 
 def var_create(varlist, name):
     cfg = vars_settings.get(name)
@@ -407,11 +315,9 @@ def var_create(varlist, name):
 
     print(f'# create variable {name}')
     var = var_template.copy()
-    var['ascii_guid'] = cfg['guid']
-    var['ascii_name'] = name
-    var['guid']       = guids.binary(cfg['guid'])
-    var['name']       = ucs16.to_ucs16(name)
-    var['attr']       = cfg['attr']
+    var['guid'] = guids.parse_str(cfg['guid'])
+    var['name'] = ucs16.from_string(name)
+    var['attr'] = cfg['attr']
     varlist[name] = var
     return var
 
@@ -428,65 +334,29 @@ def var_set_bool(varlist, name, value):
         var['data'] = b'\x00'
     var_update_time(var)
 
-def var_add_cert(varlist, name, owner, file, replace = False):
+def var_add_cert(varlist, name, owner, filename, replace = False):
     var = varlist.get(name)
     if not var:
         var = var_create(varlist, name)
-    if not var.get('siglists') or replace:
-        print(f'# init/clear {name} siglist')
-        var['siglists'] = []
+    if not var.get('sigdb') or replace:
+        print(f'# init/clear {name} sigdb')
+        var['sigdb'] = siglist.EfiSigDB()
 
-    print(f'# add {name} cert {file}')
-    with open(file, "rb") as f:
-        pem = f.read()
-    if b'-----BEGIN' in pem:
-        cert = x509.load_pem_x509_certificate(pem)
-    else:
-        cert = x509.load_der_x509_certificate(pem)
-    certdata = cert.public_bytes(serialization.Encoding.DER)
-    for c in var['siglists']:
-        if c['sigs'][0]['data'] == certdata:
-            print('# certificate already present, skipping')
-            return
-    sigs = []
-    sigs.append({
-        'ascii_guid' : owner,
-        'guid'       : guids.binary(owner),
-        'data'       : certdata,
-        'x509'       : cert,
-    })
-    var['siglists'].append({
-        'ascii_guid' : guids.EfiCertX509,
-        'guid'       : guids.binary(guids.EfiCertX509),
-        'header'     : b'',
-        'sigs'       : sigs,
-    })
-    update_data_from_siglists(var)
+    print(f'# add {name} cert {filename}')
+    var['sigdb'].add_cert(guids.parse_str(owner), filename)
+    update_data_from_sigdb(var)
     var_update_time(var)
 
-
 def var_add_dummy_dbx(varlist, owner):
     var = varlist.get('dbx')
     if var:
         return
 
-    var = var_create(varlist, 'dbx')
-    var['siglists'] = []
-
     print("# add dummy dbx entry")
-    sigs = []
-    sigs.append({
-        'ascii_guid' : owner,
-        'guid'       : guids.binary(owner),
-        'data'       : hashlib.sha256(b'').digest(),
-    })
-    var['siglists'].append({
-        'ascii_guid' : guids.EfiCertSha256,
-        'guid'       : guids.binary(guids.EfiCertSha256),
-        'header'     : b'',
-        'sigs'       : sigs,
-    })
-    update_data_from_siglists(var)
+    var = var_create(varlist, 'dbx')
+    var['sigdb'] = siglist.EfiSigDB()
+    var['sigdb'].add_dummy(guids.parse_str(owner))
+    update_data_from_sigdb(var)
     var_update_time(var)
 
 
@@ -528,6 +398,14 @@ def main():
                       action = 'append', type = 'string',
                       help = 'delete variable VAR, can be specified multiple times',
                       metavar = 'VAR')
+    parser.add_option('--set-true', dest = 'set_true',
+                      action = 'append', type = 'string',
+                      help = 'set variable VAR to true, can be specified multiple times',
+                      metavar = 'VAR')
+    parser.add_option('--set-false', dest = 'set_false',
+                      action = 'append', type = 'string',
+                      help = 'set variable VAR to false, can be specified multiple times',
+                      metavar = 'VAR')
     parser.add_option('--set-pk', dest = 'pk',  nargs = 2,
                       help = 'set PK to x509 cert, loaded in pem format ' +
                       'from FILE and with owner GUID',
@@ -567,17 +445,25 @@ def main():
         print("ERROR: no input file specified (try -h for help)")
         sys.exit(1)
 
-    print(f'"# reading varstore from {options.input}')
+    print(f'# reading varstore from {options.input}')
     with open(options.input, "rb") as f:
         infile = f.read()
 
     (start, end) = parse_volume(options.input, infile)
     print(f'var store range: 0x{start:x} -> 0x{end:x}')
-    varlist = parse_vars(infile, start, end, options.extract)
+    varlist = parse_vars(infile, start, end)
 
     if options.delete:
         vars_delete(varlist, options.delete)
 
+    if options.set_true:
+        for item in options.set_true:
+            varlist[item]['data'] = b'\x01'
+
+    if options.set_false:
+        for item in options.set_false:
+            varlist[item]['data'] = b'\x00'
+
     if options.redhat:
         platform_redhat(varlist)
         microsoft_keys(varlist)
